/*************************************************************************
 * output-prelude.c
 *
 * This module contains the output mechanism for PADS to send IDMEF alerts
 * via the prelude library
 *
 * Copyright (C) 2008 Steve Grubb <sgrubb@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************/
#include "output-prelude.h"
#include "util.h"
#include "pads.h"	// For argc/v and gc
#include <arpa/inet.h>
#include <libprelude/prelude.h>

#define ANALYZER_MODEL "pads"
#define ANALYZER_CLASS "PVS"
#define ANALYZER_MANUFACTURER "http://passive.sourceforge.net"
#define PRELUDE_FAIL_CHECK  if (ret < 0) goto err;
static prelude_client_t *client = NULL;
static char *profile = NULL;

static int setup_analyzer(idmef_analyzer_t *analyzer)
{
        int ret;
        prelude_string_t *string;

        ret = idmef_analyzer_new_model(analyzer, &string);
        PRELUDE_FAIL_CHECK;
        prelude_string_set_dup(string, ANALYZER_MODEL);

        ret = idmef_analyzer_new_class(analyzer, &string);
        PRELUDE_FAIL_CHECK;
        prelude_string_set_dup(string, ANALYZER_CLASS);

        ret = idmef_analyzer_new_manufacturer(analyzer, &string);
        PRELUDE_FAIL_CHECK;
        prelude_string_set_dup(string, ANALYZER_MANUFACTURER);

        ret = idmef_analyzer_new_version(analyzer, &string);
        PRELUDE_FAIL_CHECK;
        prelude_string_set_dup(string, PACKAGE_VERSION);

        return 0;

 err:
        log_message("%s: IDMEF error: %s.\n",
                prelude_strsource(ret), prelude_strerror(ret));

        return -1;
}

static int init_prelude(void)
{
        int ret;
        prelude_client_flags_t flags;

        ret = prelude_thread_init(NULL);
        ret = prelude_init(&prog_argc, prog_argv);
        if (ret < 0) {
                log_message("Unable to initialize the Prelude library: %s.\n",
                        prelude_strerror(ret));
                return -1;
        }
        ret = prelude_client_new(&client, profile ? profile : ANALYZER_MODEL);
        if (!client) {
                log_message("Unable to create a prelude client object: %s.\n",
                        prelude_strerror(ret));
                return -1;
        }
        ret = setup_analyzer(prelude_client_get_analyzer(client));
        if (ret < 0) {
                log_message("Unable to setup analyzer: %s\n",
                        prelude_strerror(ret));

                prelude_client_destroy(client,
                                        PRELUDE_CLIENT_EXIT_STATUS_FAILURE);
                return -1;
        }
        flags = prelude_client_get_flags(client);
        flags |= PRELUDE_CLIENT_FLAGS_ASYNC_TIMER;
        ret = prelude_client_set_flags(client, flags);
        if (ret < 0) {
                log_message("Unable to set prelude client flags: %s\n",
                        prelude_strerror(ret));

                prelude_client_destroy(client,
                                        PRELUDE_CLIENT_EXIT_STATUS_FAILURE);
                return -1;
        }
        ret = prelude_client_start(client);
        if (ret < 0) {
                log_message("Unable to start prelude client: %s\n",
                        prelude_strerror(ret));

                prelude_client_destroy(client,
                                        PRELUDE_CLIENT_EXIT_STATUS_FAILURE);
                return -1;
        }
        return 0;
}

static int new_alert_common(idmef_message_t **idmef, idmef_alert_t **alert)
{
        int ret;
        idmef_time_t *dt_time, *cr_time;

        ret = idmef_message_new(idmef);
        PRELUDE_FAIL_CHECK;

        ret = idmef_message_new_alert(*idmef, alert);
        PRELUDE_FAIL_CHECK;

        idmef_alert_set_analyzer(*alert,
                        idmef_analyzer_ref(prelude_client_get_analyzer(client)),
                        IDMEF_LIST_PREPEND);

        // Set time this was created
        ret = idmef_time_new_from_gettimeofday(&dt_time);
        PRELUDE_FAIL_CHECK;
        ret = idmef_time_clone(dt_time, &cr_time);
        PRELUDE_FAIL_CHECK;
        idmef_alert_set_detect_time(*alert, dt_time);
        idmef_alert_set_create_time(*alert, cr_time);

        return 0;
 err:
        syslog(LOG_ERR, "%s: IDMEF error: %s.\n",
               prelude_strsource(ret), prelude_strerror(ret));
        idmef_message_destroy(*idmef);
        return -1;
}

static int set_classification(idmef_alert_t *alert, const char *text)
{
        int ret;
        idmef_classification_t *classification;
        prelude_string_t *str;

        ret = idmef_alert_new_classification(alert, &classification);
        PRELUDE_FAIL_CHECK;
        ret = prelude_string_new(&str);
        PRELUDE_FAIL_CHECK;
        ret = prelude_string_set_ref(str, text);
        PRELUDE_FAIL_CHECK;
        idmef_classification_set_text(classification, str);

        return 0;
 err:
        return -1;
}

static int do_assessment(idmef_alert_t *alert,
		idmef_impact_severity_t severity,
		idmef_impact_type_t type, const char *descr)
{
        int ret;
        idmef_assessment_t *assessment;
        idmef_impact_t *impact;

        ret = idmef_alert_new_assessment(alert, &assessment);
        PRELUDE_FAIL_CHECK;
        ret = idmef_assessment_new_impact(assessment, &impact);
        PRELUDE_FAIL_CHECK;
        idmef_impact_set_severity(impact, severity);
        idmef_impact_set_type(impact, type);
        if (descr) {
               prelude_string_t *str;
               ret = idmef_impact_new_description(impact, &str);
               prelude_string_set_ref(str, descr);
        }
        idmef_impact_set_completion(impact, IDMEF_IMPACT_COMPLETION_SUCCEEDED);
        return 0;
 err:
        return -1;
}

// FIXME: handle ipv6 addresses
static int fill_in_node_addr(idmef_node_t *node, struct in_addr *addr)
{
        int ret;
        prelude_string_t *str;
	char buf[INET6_ADDRSTRLEN+1];

	inet_ntop(AF_INET, addr, buf, sizeof(buf));

        /* Setup the address string */
        ret = prelude_string_new(&str);
        PRELUDE_FAIL_CHECK;
        ret = prelude_string_set_dup(str, buf);
        PRELUDE_FAIL_CHECK;

        /* Now record IP address */
        idmef_address_t *my_addr;
        ret = idmef_address_new(&my_addr);
        PRELUDE_FAIL_CHECK;
        idmef_address_set_category(my_addr, IDMEF_ADDRESS_CATEGORY_IPV4_ADDR);
        idmef_address_set_address(my_addr, str);
        idmef_node_set_address(node, my_addr, 0);

        return 0;
 err:
        return -1;
}

// FIXME: handle ipv6 addresses
static int fill_in_node_mac(idmef_node_t *node, ArpAsset *a)
{
        int ret;
        prelude_string_t *str, *mac_str;
	char buf[INET6_ADDRSTRLEN+1];

	inet_ntop(AF_INET, &a->ip_addr, buf, sizeof(buf));

        /* Setup the address string */
        ret = prelude_string_new(&str);
        PRELUDE_FAIL_CHECK;
        ret = prelude_string_set_dup(str, buf);
        PRELUDE_FAIL_CHECK;

        /* Now record IP address */
        idmef_address_t *my_addr;
        ret = idmef_address_new(&my_addr);
        PRELUDE_FAIL_CHECK;
        idmef_address_set_category(my_addr, IDMEF_ADDRESS_CATEGORY_IPV4_ADDR);
        idmef_address_set_address(my_addr, str);
        idmef_node_set_address(node, my_addr, 0);

        /* Now record MAC address */
        idmef_address_t *mac_addr;
        ret = idmef_address_new(&mac_addr);
        PRELUDE_FAIL_CHECK;
        ret = prelude_string_new(&mac_str);
        PRELUDE_FAIL_CHECK;
        idmef_address_set_category(mac_addr, IDMEF_ADDRESS_CATEGORY_MAC);
        ret = prelude_string_set_dup(mac_str, hex2mac(a->mac_addr));
        PRELUDE_FAIL_CHECK;
        idmef_address_set_address(mac_addr, mac_str);
        idmef_node_set_address(node, mac_addr, 0);
        return 0;
 err:
        return -1;
}


static int asset_alert(Asset *a, idmef_message_t *idmef,
                idmef_alert_t *alert, const char *msg,
                idmef_impact_severity_t severity, char *descr)
{
        int ret;
        idmef_source_t *source;
        idmef_target_t *target;
	idmef_node_t *node, *node2;
	idmef_service_t *service;
        idmef_impact_type_t impact;
        prelude_string_t *name, *str;
	idmef_additional_data_t *data;

        /* Fill in information about the event's source */
        ret = idmef_alert_new_source(alert, &source, -1);
        PRELUDE_FAIL_CHECK;
	ret = idmef_source_new_node(source, &node);
        PRELUDE_FAIL_CHECK;
	idmef_node_set_category(node, IDMEF_NODE_CATEGORY_UNKNOWN);
	ret = fill_in_node_addr(node, &a->ip_addr);
	PRELUDE_FAIL_CHECK;

	// Describe the service
	ret = idmef_source_new_service(source, &service);
	PRELUDE_FAIL_CHECK;
	idmef_service_set_iana_protocol_number(service, a->proto);
	idmef_service_set_ip_version(service, 4);
	idmef_service_set_port(service, ntohs(a->port));
	ret = prelude_string_new(&name);
	PRELUDE_FAIL_CHECK;
	ret = prelude_string_set_dup(name, (char *)bdata(a->service));
	PRELUDE_FAIL_CHECK;
        idmef_service_set_name(service, name);
	if (bdata(a->application) != NULL) {
		ret = idmef_alert_new_additional_data(alert, &data,
			IDMEF_LIST_APPEND);
		PRELUDE_FAIL_CHECK;
		ret = idmef_additional_data_new_meaning(data, &str);
		PRELUDE_FAIL_CHECK;
		prelude_string_set_ref(str, "Application");
		idmef_additional_data_set_type(data,
				IDMEF_ADDITIONAL_DATA_TYPE_STRING);
		idmef_additional_data_set_string_dup(data,
				(char *)bdata(a->application));
	}

        /* Fill in information about the target of the event */
        ret = idmef_alert_new_target(alert, &target, -1);
        PRELUDE_FAIL_CHECK;

	/* FIXME: I assume that the target is the whole network */
	ret = idmef_target_new_node(target, &node2);
        PRELUDE_FAIL_CHECK;
	idmef_node_set_category(node, IDMEF_NODE_CATEGORY_UNKNOWN);

        /* Describe event */
        ret = set_classification(alert, msg);
        PRELUDE_FAIL_CHECK;

        /* Assess impact */
        impact = IDMEF_IMPACT_TYPE_OTHER;
        ret = do_assessment(alert, severity, impact, descr);
        PRELUDE_FAIL_CHECK;

        prelude_client_send_idmef(client, idmef);
        idmef_message_destroy(idmef);

        return 0;
 err:
        syslog(LOG_ERR, "asset_alert: IDMEF error: %s.\n",
                prelude_strerror(ret));
        idmef_message_destroy(idmef);
        return -1;
}


static int arp_alert(ArpAsset *a, idmef_message_t *idmef,
                idmef_alert_t *alert, const char *msg,
                idmef_impact_severity_t severity, char *descr)
{
        int ret;
        idmef_source_t *source;
        idmef_target_t *target;
	idmef_node_t *node, *node2;
        idmef_impact_type_t impact;
        prelude_string_t *str;
	idmef_additional_data_t *data;

        /* Fill in information about the event's source */
        ret = idmef_alert_new_source(alert, &source, -1);
        PRELUDE_FAIL_CHECK;
	ret = idmef_source_new_node(source, &node);
        PRELUDE_FAIL_CHECK;
	idmef_node_set_category(node, IDMEF_NODE_CATEGORY_UNKNOWN);
	ret = fill_in_node_mac(node, a);
	PRELUDE_FAIL_CHECK;

	if (bdata(a->mac_resolved) != NULL) {
		ret = idmef_alert_new_additional_data(alert, &data,
			IDMEF_LIST_APPEND);
		PRELUDE_FAIL_CHECK;
		ret = idmef_additional_data_new_meaning(data, &str);
		PRELUDE_FAIL_CHECK;
		prelude_string_set_ref(str, "Vendor");
		idmef_additional_data_set_type(data,
				IDMEF_ADDITIONAL_DATA_TYPE_STRING);
		idmef_additional_data_set_string_dup(data,
				(char *)bdata(a->mac_resolved));
	}

        /* Fill in information about the target of the event */
        ret = idmef_alert_new_target(alert, &target, -1);
        PRELUDE_FAIL_CHECK;

	/* FIXME: I assume that the target is the whole network */
	ret = idmef_target_new_node(target, &node2);
        PRELUDE_FAIL_CHECK;
	idmef_node_set_category(node, IDMEF_NODE_CATEGORY_UNKNOWN);

        /* Describe event */
        ret = set_classification(alert, msg);
        PRELUDE_FAIL_CHECK;

        /* Assess impact */
        impact = IDMEF_IMPACT_TYPE_OTHER;
        ret = do_assessment(alert, severity, impact, descr);
        PRELUDE_FAIL_CHECK;

        prelude_client_send_idmef(client, idmef);
        idmef_message_destroy(idmef);

        return 0;
 err:
        syslog(LOG_ERR, "arp_alert: IDMEF error: %s.\n",
                prelude_strerror(ret));
        idmef_message_destroy(idmef);
        return -1;
}

/* ----------------------------------------------------------
 * FUNCTION	: setup_output_screen
 * DESCRIPTION	: This function will register the output
 *		: plugin.
 * INPUT	: None!
 * RETURN	: 0 - Success
 *		: -1 - Error
 * ---------------------------------------------------------- */
int
setup_output_prelude (void)
{
    OutputPlugin *plugin;


    /* Allocate and setup plugin data record. */
    plugin = (OutputPlugin*)malloc(sizeof(OutputPlugin));
    plugin->name = bfromcstr("prelude");
    plugin->init = init_output_prelude;
    plugin->print_asset = print_asset_prelude;
    plugin->print_arp = print_arp_asset_prelude;
    plugin->print_stat = NULL;
    plugin->end = end_output_prelude;

    /* Register plugin with input module. */
    if ((register_output_plugin(plugin)) == -1) {
	if (plugin != NULL)
	    free(plugin);
	log_message("warning:  'register_output_plugin' in function 'setup_output_prelude' failed.");
	return -1;
    }

    return 0;
}

/* ----------------------------------------------------------
 * FUNCTION	: init_output_prelude
 * DESCRIPTION	: This output module will initialize the prelude
 *		: subsystem.
 * INPUT	: None
 * RETURN	: 0 - Success
 *		: -1 - Error
 * --------------------------------------------------------- */
int
init_output_prelude (bstring args)
{
    verbose_message("Initializing PRELUDE output processor.");
    profile = strdup((char *)bdata(args));

    if (init_prelude() < 0)
	return -1;

    return 0;
}

/* ----------------------------------------------------------
 * FUNCTION	: print_asset_prelude
 * DESCRIPTION	: This function will send asset events to prelude
 * INPUT	: 0 - Port
 *		: 1 - IP Address
 *		: 2 - Service
 *		: 3 - Application
 * RETURN	: 0 - Success
 *		: -1 - Error
 * ---------------------------------------------------------- */
int
print_asset_prelude (Asset *rec)
{
    idmef_message_t *idmef;
    idmef_alert_t *alert;

    if (new_alert_common(&idmef, &alert) >= 0){ 
	asset_alert(rec, idmef, alert, 
		"New Network Asset Detected", IDMEF_IMPACT_SEVERITY_LOW,
		"A service that was not previously known has been used.");
    }
    return 0;
}


/* ----------------------------------------------------------
 * FUNCTION	: print_arp_asset_prelude
 * DESCRIPTION	: This function will send ARP asset events to prelude
 * INPUT	: 0 - IP Address
 *		: 1 - MAC Address
 * RETURN	: 0 - Success
 *		: -1 - Error
 * ---------------------------------------------------------- */
int
print_arp_asset_prelude (ArpAsset *rec)
{
    idmef_message_t *idmef;
    idmef_alert_t *alert;

    if (new_alert_common(&idmef, &alert) >= 0){ 
	arp_alert(rec, idmef, alert, 
		"New MAC Address Detected", IDMEF_IMPACT_SEVERITY_INFO,
		"A network card that was not previously known has been used.");
    }
    return 0;
}

/* ----------------------------------------------------------
 * FUNCTION	: end_output_prelude
 * DESCRIPTION	: This function will free the resources used
 *		: by the prelude output module.
 * INPUT	: None!
 * RETURN	: None!
 * ---------------------------------------------------------- */
int
end_output_prelude ()
{
    verbose_message("Ending PRELUDE Output Processor.");

    if (client)
        prelude_client_destroy(client, PRELUDE_CLIENT_EXIT_STATUS_SUCCESS);
    prelude_deinit();
    free(profile);

    return 0;
}

